@@ -1,6 +1,5 @@
 """Allows the package to be run with `python3 -m ubuntu_image`."""
 
-
 import os
 import sys
 import logging
@@ -8,27 +7,15 @@ import argparse
 
 from contextlib import suppress
 from pickle import dump, load
-from pkg_resources import resource_string as resource_bytes
+from ubuntu_image import __version__
 from ubuntu_image.builder import ModelAssertionBuilder
 from ubuntu_image.helpers import as_size
 from ubuntu_image.i18n import _
+from ubuntu_image.parser import GadgetSpecificationError
 
 
 _logger = logging.getLogger('ubuntu-image')
 
-# Try to get the version number, which will be different if we're living in a
-# snap world or a deb.  Actually, I'd prefer to not even have the -NubuntuY
-# version string when we're running from source, but that's trickier, so don't
-# worry about it.
-__version__ = os.environ.get('SNAP_VERSION')
-if __version__ is None:                                      # pragma: nocover
-    try:
-        __version__ = resource_bytes(
-            'ubuntu_image', 'version.txt').decode('utf-8')
-    except FileNotFoundError:
-        # Probably, setup.py hasn't been run yet to generate the version.txt.
-        __version__ = 'dev'
-
 
 PROGRAM = 'ubuntu-image'
 
@@ -169,9 +156,19 @@ def main(argv=None):
             state_machine.run_until(args.until)
         else:
             list(state_machine)
+    except GadgetSpecificationError as error:
+        if args.debug:
+            _logger.exception('gadget.yaml parse error')
+        else:
+            _logger.error('gadget.yaml parse error: {}'.format(error))
+            _logger.error('Use --debug for more information')
     except:
         _logger.exception('Crash in state machine')
         return 1
+    # It's possible that the state machine didn't crash, but it still didn't
+    # complete successfully.  For example, if `snap prepare-image` failed.
+    if state_machine.exitcode != 0:
+        return state_machine.exitcode
     # Everything's done, now handle saving state if necessary.
     if pickle_file is not None:
         with open(pickle_file, 'wb') as fp:
