@@ -2,12 +2,16 @@
 
 import os
 import shutil
+import logging
 
+from contextlib import ExitStack
 from pkg_resources import resource_filename
 from ubuntu_image.builder import ModelAssertionBuilder
+from unittest.mock import patch
 
 
 class XXXModelAssertionBuilder(ModelAssertionBuilder):
+    exitcode = 0
     gadget_yaml = 'gadget.yaml'
 
     # We need this class because the current gadget snap we get from the store
@@ -19,8 +23,8 @@ class XXXModelAssertionBuilder(ModelAssertionBuilder):
         meta_dir = os.path.join(gadget_dir, 'meta')
         os.makedirs(meta_dir, exist_ok=True)
         shutil.copy(
-            resource_filename('ubuntu_image.tests.data', 'gadget.yaml'),
-            os.path.join(meta_dir, self.gadget_yaml))
+            resource_filename('ubuntu_image.tests.data', self.gadget_yaml),
+            os.path.join(meta_dir, 'gadget.yaml'))
         shutil.copy(
             resource_filename('ubuntu_image.tests.data', 'grubx64.efi'),
             os.path.join(gadget_dir, 'grubx64.efi'))
@@ -58,3 +62,28 @@ class EarlyExitLeaveATraceAssertionBuilder(XXXModelAssertionBuilder):
         # have something to positively test.
         with open(os.path.join(self.workdir, 'success'), 'w'):
             pass
+
+
+class LogCapture:
+    def __init__(self):
+        self.logs = []
+        self._resources = ExitStack()
+
+    def capture(self, *args, **kws):
+        level, fmt, fmt_args = args
+        self.logs.append((level, fmt % fmt_args))
+        # Was .exception() called?
+        exc_info = kws.pop('exc_info', None)
+        assert len(kws) == 0, kws
+        if exc_info:
+            self.logs.append('IMAGINE THE TRACEBACK HERE')
+
+    def __enter__(self):
+        log = logging.getLogger('ubuntu-image')
+        self._resources.enter_context(patch.object(log, '_log', self.capture))
+        return self
+
+    def __exit__(self, *exception):
+        self._resources.close()
+        # Don't suppress any exceptions.
+        return False
