@@ -7,16 +7,55 @@ from contextlib import ExitStack
 from io import StringIO
 from pickle import load
 from pkg_resources import resource_filename
+from subprocess import CalledProcessError
 from tempfile import TemporaryDirectory
-from ubuntu_image.__main__ import main
+from types import SimpleNamespace
+from ubuntu_image.__main__ import main, parseargs
+from ubuntu_image.helpers import GiB, MiB
 from ubuntu_image.testing.helpers import (
     CrashingModelAssertionBuilder, DoNothingBuilder,
     EarlyExitLeaveATraceAssertionBuilder, EarlyExitModelAssertionBuilder,
-    XXXModelAssertionBuilder)
+    LogCapture, XXXModelAssertionBuilder)
+from ubuntu_image.testing.nose import NosePlugin
 from unittest import TestCase, skipIf
 from unittest.mock import call, patch
 
 
+# For forcing a test failure.
+def check_returncode(*args, **kws):
+    raise CalledProcessError(1, 'failing command')
+
+
+class BadGadgetModelAssertionBuilder(XXXModelAssertionBuilder):
+    gadget_yaml = 'bad-gadget.yaml'
+
+
+class TestParseArgs(TestCase):
+    def test_image_size_option_bytes(self):
+        args = parseargs(['--image-size', '45', 'model.assertion'])
+        self.assertEqual(args.image_size, 45)
+        self.assertEqual(args.given_image_size, '45')
+
+    def test_image_size_option_suffixes(self):
+        args = parseargs(['--image-size', '45G', 'model.assertion'])
+        self.assertEqual(args.image_size, GiB(45))
+        self.assertEqual(args.given_image_size, '45G')
+        args = parseargs(['--image-size', '45M', 'model.assertion'])
+        self.assertEqual(args.image_size, MiB(45))
+        self.assertEqual(args.given_image_size, '45M')
+
+    def test_image_size_option_invalid(self):
+        # These errors will output to stderr, but that just clouds the test
+        # output, so suppress it.
+        with patch('argparse._sys.stderr'):
+            self.assertRaises(SystemExit,
+                              parseargs,
+                              ['--image-size', '45Q', 'model.assertion'])
+            self.assertRaises(SystemExit,
+                              parseargs,
+                              ['--image-size', 'BIG', 'model.assertion'])
+
+
 class TestMain(TestCase):
     def setUp(self):
         super().setUp()
@@ -77,6 +116,36 @@ class TestMain(TestCase):
             self.assertEqual(
                 mock.call_args_list[-1], call('Crash in state machine'))
 
+    def test_state_machine_snap_command_fails(self):
+        # The `snap prepare-image` command fails and main exits with non-zero.
+        #
+        # This tests needs to run the actual snap() helper function, not
+        # the testsuite-wide mock.  This is appropriate since we're
+        # mocking it ourselves here.
+        if NosePlugin.snap_mocker is not None:
+            NosePlugin.snap_mocker.patcher.stop()
+            self._resources.callback(NosePlugin.snap_mocker.patcher.start)
+        self._resources.enter_context(patch(
+            'ubuntu_image.helpers.subprocess_run',
+            return_value=SimpleNamespace(
+                returncode=1,
+                stdout='command stdout',
+                stderr='command stderr',
+                check_returncode=check_returncode,
+                )))
+        self._resources.enter_context(LogCapture())
+        self._resources.enter_context(patch(
+            'ubuntu_image.__main__.ModelAssertionBuilder',
+            XXXModelAssertionBuilder))
+        workdir = self._resources.enter_context(TemporaryDirectory())
+        imgfile = os.path.join(workdir, 'my-disk.img')
+        code = main(('--until', 'prepare_filesystems',
+                     '--channel', 'edge',
+                     '--workdir', workdir,
+                     '--output', imgfile,
+                     'model.assertion'))
+        self.assertEqual(code, 1)
+
 
 class TestMainWithModel(TestCase):
     def setUp(self):
@@ -182,3 +251,49 @@ class TestMainWithModel(TestCase):
         self.assertFalse(os.path.exists(os.path.join(workdir, 'success')))
         main(('--workdir', workdir, '--resume'))
         self.assertTrue(os.path.exists(os.path.join(workdir, 'success')))
+
+
+class TestMainWithBadGadget(TestCase):
+    def setUp(self):
+        super().setUp()
+        self._resources = ExitStack()
+        self.addCleanup(self._resources.close)
+        self.model_assertion = resource_filename(
+            'ubuntu_image.tests.data', 'model.assertion')
+
+    @skipIf('UBUNTU_IMAGE_TESTS_NO_NETWORK' in os.environ,
+            'Cannot run this test without network access')
+    def test_bad_gadget_log(self):
+        log = self._resources.enter_context(LogCapture())
+        workdir = self._resources.enter_context(TemporaryDirectory())
+        self._resources.enter_context(patch(
+            'ubuntu_image.__main__.ModelAssertionBuilder',
+            BadGadgetModelAssertionBuilder))
+        main(('--channel', 'edge',
+              '--workdir', workdir,
+              self.model_assertion))
+        self.assertEqual(log.logs, [
+            (logging.ERROR, 'gadget.yaml parse error: '
+                            'GUID structure type with non-GPT schema'),
+            (logging.ERROR, 'Use --debug for more information')
+            ])
+
+    @skipIf('UBUNTU_IMAGE_TESTS_NO_NETWORK' in os.environ,
+            'Cannot run this test without network access')
+    def test_bad_gadget_debug_log(self):
+        log = self._resources.enter_context(LogCapture())
+        workdir = self._resources.enter_context(TemporaryDirectory())
+        self._resources.enter_context(patch(
+            'ubuntu_image.__main__.ModelAssertionBuilder',
+            BadGadgetModelAssertionBuilder))
+        main(('--channel', 'edge',
+              '--debug',
+              '--workdir', workdir,
+              self.model_assertion))
+        self.assertEqual(log.logs, [
+            (logging.ERROR, 'uncaught exception in state machine step: '
+                            '[2] load_gadget_yaml'),
+            'IMAGINE THE TRACEBACK HERE',
+            (logging.ERROR, 'gadget.yaml parse error'),
+            'IMAGINE THE TRACEBACK HERE',
+            ])
